# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sdc_scissor',
 'sdc_scissor.feature_extraction_api',
 'sdc_scissor.machine_learning_api',
 'sdc_scissor.obstacle_api',
 'sdc_scissor.sdc_prioritizer',
 'sdc_scissor.sdc_prioritizer.code.python',
 'sdc_scissor.sdc_prioritizer.datasets.fullroad.BeamNG_AI.BeamNG_RF_1_5',
 'sdc_scissor.sdc_prioritizer.testPrioritization',
 'sdc_scissor.sdc_prioritizer.testPrioritization.crossover',
 'sdc_scissor.sdc_prioritizer.testPrioritization.mutation',
 'sdc_scissor.sdc_prioritizer.testPrioritization.problem',
 'sdc_scissor.simulator_api',
 'sdc_scissor.testing_api',
 'sdc_scissor.testing_api.test_generators.ambiegen',
 'sdc_scissor.testing_api.test_generators.ambiegen.Utils',
 'sdc_scissor.testing_api.test_generators.frenetic',
 'sdc_scissor.testing_api.test_generators.frenetic.src',
 'sdc_scissor.testing_api.test_generators.frenetic.src.generators',
 'sdc_scissor.testing_api.test_generators.frenetic.src.utils',
 'sdc_scissor.testing_api.test_generators.frenetic_v',
 'sdc_scissor.testing_api.test_generators.frenetic_v.src',
 'sdc_scissor.testing_api.test_generators.frenetic_v.src.generators',
 'sdc_scissor.testing_api.test_generators.frenetic_v.src.utils']

package_data = \
{'': ['*'],
 'sdc_scissor.sdc_prioritizer': ['code/*',
                                 'code/r-script/*',
                                 'datasets/fullroad/*',
                                 'datasets/fullroad/BeamNG_AI/*',
                                 'datasets/fullroad/BeamNG_AI/BeamNG_RF_1/*',
                                 'datasets/fullroad/BeamNG_AI/BeamNG_RF_2/*',
                                 'datasets/fullroad/Driver_AI/*',
                                 'docker_scripts/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'Shapely>=1.8.1,<2.0.0',
 'beamngpy==1.22',
 'click>=8.0.4,<9.0.0',
 'pandas>=1.4.1,<2.0.0',
 'pymoo==0.4.2.2',
 'scikit-learn>=1.0.2,<2.0.0']

entry_points = \
{'console_scripts': ['sdc-scissor = sdc_scissor.cli:cli']}

setup_kwargs = {
    'name': 'sdc-scissor',
    'version': '2.0.2',
    'description': 'A cost-effective test selection for self-driving cars in virtual environments',
    'long_description': '# SDC-Scissor\n[![License: GPL v3](https://img.shields.io/badge/License-GPLv3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)\n[![](https://github.com/ChristianBirchler/sdc-scissor/actions/workflows/ci.yml/badge.svg)](https://github.com/ChristianBirchler/sdc-scissor/actions/workflows/ci.yml)\n[![CD](https://github.com/ChristianBirchler/sdc-scissor/actions/workflows/cd.yml/badge.svg)](https://github.com/ChristianBirchler/sdc-scissor/actions/workflows/cd.yml)\n[![](https://readthedocs.org/projects/sdc-scissor/badge)](https://sdc-scissor.readthedocs.io)\n[![](https://img.shields.io/badge/code%20style-black-000000.svg)](https://black.readthedocs.io/)\n[![](https://sonarcloud.io/api/project_badges/measure?project=ChristianBirchler_sdc-scissor&metric=alert_status)](https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor)\n[![](https://sonarcloud.io/api/project_badges/measure?project=ChristianBirchler_sdc-scissor&metric=ncloc)](https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor)\n[![](https://sonarcloud.io/api/project_badges/measure?project=ChristianBirchler_sdc-scissor&metric=coverage)](https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor)\n[![](https://sonarcloud.io/api/project_badges/measure?project=ChristianBirchler_sdc-scissor&metric=sqale_index)](https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor)\n[![](https://sonarcloud.io/api/project_badges/measure?project=ChristianBirchler_sdc-scissor&metric=reliability_rating)](https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor)\n[![](https://sonarcloud.io/api/project_badges/measure?project=ChristianBirchler_sdc-scissor&metric=duplicated_lines_density)](https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor)\n[![](https://sonarcloud.io/api/project_badges/measure?project=ChristianBirchler_sdc-scissor&metric=vulnerabilities)](https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor)\n[![](https://sonarcloud.io/api/project_badges/measure?project=ChristianBirchler_sdc-scissor&metric=bugs)](https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor)\n[![](https://sonarcloud.io/api/project_badges/measure?project=ChristianBirchler_sdc-scissor&metric=security_rating)](https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor)\n[![](https://sonarcloud.io/api/project_badges/measure?project=ChristianBirchler_sdc-scissor&metric=sqale_rating)](https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor)\n[![](https://sonarcloud.io/api/project_badges/measure?project=ChristianBirchler_sdc-scissor&metric=code_smells)](https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor)\n\n## A Tool for Cost-effective Simulation-based Test Selection in Self-driving Cars Software\n<div style="text-align: center;">\n<a href="https://github.com/ChristianBirchler/sdc-scissor">\n<img src="https://raw.githubusercontent.com/ChristianBirchler/sdc-scissor/main/docs/images/github_logo_icon.png">\n</a>\n<a href="https://sonarcloud.io/summary/overall?id=ChristianBirchler_sdc-scissor">\n<img src="https://sonarcloud.io/images/project_badges/sonarcloud-black.svg">\n</a>\n</div>\n\nSDC-Scissor is a tool that let you test self-driving cars more efficiently in simulation. It uses a machine-learning\napproach to select only relevant test scenarios so that the testing process is faster. Furthermore, the selected tests\nare diverse and try to challenge the car with corner cases.\n\nFurthermore, this repository contains also code for test multi-objective test case prioritization with an evolutionary\ngenetic search algorithm. If you are interested in prioritizing test cases, then you should read the dedicated\n[README.md](https://github.com/ChristianBirchler/sdc-scissor/blob/main/sdc_scissor/sdc_prioritizer/testPrioritization/README.md)\nfor this. If you use the prioritization technique then also cite the papers from the reference section!\n\n## Support\nJoin the community on [Slack](https://slack.com/) or [Discord](https://discord.com/) to get support of developers and\nusers. Just click on one of the following links:\n\n* [https://join.slack.com/t/sdc-scissor/shared_invite/zt-1aikrj1uu-Dz0a9BE1AQ4GQp4A3Bm5og](https://join.slack.com/t/sdc-scissor/shared_invite/zt-1aikrj1uu-Dz0a9BE1AQ4GQp4A3Bm5og)\n* [https://discord.gg/AaYnBS6s7E](https://discord.gg/AaYnBS6s7E)\n\n## Documentation\n[![](https://raw.githubusercontent.com/ChristianBirchler/sdc-scissor/main/docs/images/readthedocs.png)](https://sdc-scissor.readthedocs.io/en/latest/)\n\n## License\n```{code-block} text\nSDC-Scissor tool for cost-effective simulation-based test selection\nin self-driving cars software.\nCopyright (C) 2022  Christian Birchler\n\nThis program is free software: you can redistribute it and/or modify\nit under the terms of the GNU General Public License as published by\nthe Free Software Foundation, either version 3 of the License, or\n(at your option) any later version.\n\nThis program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\nGNU General Public License for more details.\n\nYou should have received a copy of the GNU General Public License\nalong with this program.  If not, see <https://www.gnu.org/licenses/>.\n```\n\nThe software we developed is distributed under GNU GPL license. See the\n[LICENSE.md](https://github.com/ChristianBirchler/sdc-scissor/blob/main/LICENSE.md) file.\n\n## References\nIf you use this tool in your research, please cite the following papers:\n\n* Christian Birchler, Nicolas Ganz, Sajad Khatiri, Alessio Gambi, and Sebastiano Panichella. 2022. Cost-effective\nSimulationbased Test Selection in Self-driving Cars Software with SDC-Scissor. In 2022 IEEE 29th International\nConference on Software Analysis, Evolution and Reengineering (SANER), IEEE.\n\n* Christian Birchler, Sajad Khatiri, Pouria Derakhshanfar, Sebastiano Panichella, and Annibale Panichella. 2022.\nSingle and Multi-objective Test Cases Prioritization for Self-driving Cars in Virtual Environments. ACM Transactions on\nSoftware Engineering and Methodology (TOSEM) (2022). DOI: to appear\n\n```{code-block} bibtex\n@inproceedings{Birchler2022Cost,\n  author={Birchler, Christian and Ganz, Nicolas and Khatiri, Sajad and Gambi, Alessio, and Panichella, Sebastiano},\n  booktitle={2022 IEEE 29th International Conference on Software Analysis, Evolution and Reengineering (SANER)},\n  title={Cost-effective Simulationbased Test Selection in Self-driving Cars Software with SDC-Scissor},\n  year={2022},\n  doi={to appear}\n}\n\n@article{Birchler2022Single,\n  author={Birchler, Christian and Khatiri, Sajad and Derakhshanfar, Pouria and Panichella, Sebastiano and Panichella, Annibale},\n  title={Single and Multi-objective Test Cases Prioritization for Self-driving Cars in Virtual Environments},\n  year={2022},\n  publisher={Association for Computing Machinery},\n  journal={ACM Transactions on Software Engineering and Methodology (TOSEM)},\n  doi={to appear}\n}\n```\n\n## Contacts\n* Christian Birchler\n    * Zurich University of Applied Science (ZHAW), Switzerland - birc@zhaw.ch\n* Nicolas Ganz\n    * Zurich University of Applied Science (ZHAW), Switzerland - gann@zhaw.ch\n* Sajad Khatiri\n    * Zurich University of Applied Science (ZHAW), Switzerland - mazr@zhaw.ch\n* Dr. Alessio Gambi\n    * Passau University, Germany - alessio.gambi@uni-passau.de\n* Dr. Sebastiano Panichella\n    * Zurich University of Applied Science (ZHAW), Switzerland - panc@zhaw.ch\n',
    'author': 'Christian Birchler',
    'author_email': 'birchler.chr@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ChristianBirchler/sdc-scissor',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from pymoo.core.problem import Problem
import multiprocessing as mp


def ff_eval(x):
    global costs
    global distances
    f1 = 0.0
    f2 = 0.0

    for index, test_id in enumerate(x[:-1]):
        f1 += costs[test_id] / (index + 1)
        f2 -= distances[test_id][x[index + 1]] / (index + 1)
    return f1, f2


class TestPrioritizationMultiObjectiveProblem(Problem):
    generation_counter = 0

    def __init__(self, genome_length, distancesp, costsp):

        super().__init__(n_var=genome_length, n_obj=2, n_constr=0, xl=0.0, xu=genome_length)
        global costs
        costs = costsp
        global distances
        distances = distancesp

    def _evaluate(self, x, out, *args, **kwargs):

        pool1 = mp.Pool(processes=5)

        new_rows2 = pool1.map(ff_eval, x)
        pool1.terminate()
        out["F"] = new_rows2



from pymoo.model.problem import Problem


class TestCaseProblem(Problem):
    """
    Module to calculate the fitnes of the individuals
    """

    def __init__(self):
        super().__init__(n_var=1, n_obj=2, n_constr=1, elementwise_evaluation=True)

    def _evaluate(self, x, out, *args, **kwargs):
        s = x[0]
        s.get_points()  # transform the states into actual points (mutation and crossover operations are performed on states)
        s.remove_invalid_cases()
        s.eval_fitness()
        out["F"] = [s.fitness, s.novelty]
        out["G"] = 4 - s.fitness * (-1)



import math
from sdc_scissor.testing_api.test_generators.frenetic_v.src.generators.base_generator import BaseGenerator
from sdc_scissor.testing_api.test_generators.frenetic_v.src.utils import frenet

from shapely import geometry, affinity


import numpy as np
import logging as log


class BaseFrenetVGenerator(BaseGenerator):
    def __init__(
        self,
        executor=None,
        map_size=None,
        segment_length=10,
        lane_width=4,
        strict_father=False,
        store_additional_data=False,
    ):
        # Margin size w.r.t the map
        self.segment_length = segment_length
        self.lane_width = lane_width
        self.margin = 10  # to be safe... there were some out of margin with (lane_width * 2)
        self.recent_count = 0
        self.theta0 = 1.57  # theta0: The initial angle of the line. (1.57 == 90 degrees)
        super().__init__(
            executor=executor,
            map_size=map_size,
            strict_father=strict_father,
            store_additional_data=store_additional_data,
        )

    def kappas_to_road_points(self, kappas) -> np.array:
        """
        The FreneticV main algorithm.

        :param kappas: list of kappa values
        :return: road points in cartesian coordinates
        """
        # Using the bottom center of the map.
        y0 = self.margin
        x0 = self.map_size / 2
        theta0 = self.theta0
        ss = np.cumsum([self.segment_length] * len(kappas)) - self.segment_length
        # Transforming the frenet points to cartesian
        (xs, ys) = frenet.frenetv_to_cartesian(x0, y0, theta0, ss, kappas)

        return np.column_stack([xs, ys])

    def execute_frenet_test(self, kappas, method="random", parent_info={}, extra_info={}):
        extra_info["kappas"] = kappas
        road_points = self.kappas_to_road_points(kappas)

        # Pre-validation
        # 1. check self-intersection
        if self.is_likely_self_intersecting(road_points):
            return "LIKELY_SELF_INTERSECTING", None

        # 2. check if it a) fits into map-dimensions and b) if we can transform it to be inside the map
        road_points = self.reframe_road(road_points)

        if road_points is not None:
            self.recent_count += 1
            # Transforming numpy array to list of tuples
            road_points_list = list(map(tuple, road_points))
            # return self.execute_test(road_points_list, method=method, parent_info=parent_info, extra_info=extra_info)
            return road_points_list
        else:
            return "CANNOT_REFRAME", None

    def is_likely_self_intersecting(self, road_points) -> bool:
        center_line = geometry.LineString(road_points)

        left_line = center_line.parallel_offset(self.lane_width, "left")
        right_line = center_line.parallel_offset(self.lane_width, "right")

        drop_road_checks = {
            "left_complex": not left_line.is_simple,
            "center_complex": not center_line.is_simple,
            "right_complex": not right_line.is_simple,
            "left_int_right": left_line.intersects(right_line),
            "right_int_left": right_line.intersects(left_line),
            "left_ring": left_line.is_ring,
            "center_ring": center_line.is_ring,
            "right_ring": right_line.is_ring,
            "left_is_multi": left_line.geom_type != "LineString",
            "right_is_multi": right_line.geom_type != "LineString",
        }

        return any(drop_road_checks.values())

    def rotate_and_relocate_line(self, center_line, orientation):
        reoriented_center = affinity.rotate(center_line, orientation)
        reoriented_convex_hull = reoriented_center.convex_hull
        reoriented_convex_hull_coords = np.array(reoriented_convex_hull.exterior.coords)

        xoff = self.margin - min(reoriented_convex_hull_coords[:, 0])  # where's the left side?
        yoff = self.margin - min(reoriented_convex_hull_coords[:, 1])  # where's the bottom?
        new_center_line = affinity.translate(reoriented_center, xoff=xoff, yoff=yoff)
        return new_center_line

    def reframe_road(self, road_points):
        center_line = geometry.LineString(road_points)

        reduced_map_rect = geometry.Polygon(
            [
                (self.margin, self.margin),
                (self.margin, self.map_size - self.margin),
                (self.map_size - self.margin, self.map_size - self.margin),
                (self.map_size - self.margin, self.margin),
            ]
        )

        # check if the line is in the map already
        if reduced_map_rect.contains(center_line):
            log.debug("The road already fits.")
            return road_points

        # rotation and alignment

        # Aligning based on best-fit orientation of two rectangles
        # https://math.stackexchange.com/questions/3179927/determining-best-angle-of-rotation-to-fit-a-rectangle-inside-a-rectangle
        # l = geometry.LineString(rectangular_hull_coords[0:2]).length
        # b = geometry.LineString(rectangular_hull_coords[1:3]).length
        # orientation = math.acos(self.map_size / math.sqrt(l ** 2 + b ** 2)) + math.atan(b / l)
        # new_center_line = self.rotate_and_relocate_line(center_line, orientation)

        # If aligning rectangles does not always work...
        for orientation in range(0, 90, 1):
            new_center_line = self.rotate_and_relocate_line(center_line, orientation)
            if reduced_map_rect.contains(new_center_line):
                log.info("Road was relocated and rotated")
                return np.array(new_center_line.coords)

        log.info("The road could not be fit within the boundaries.")
        return None



#
import random as rm
import numpy as np
import math as m
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon
import matplotlib.pyplot as plt

import time


class Map:
    """Class that conducts transformations to vectors automatically,
    using the commads "go straight", "turn left", "turn right".
    As a result it produces a set of points corresponding to a road

    """

    def __init__(self, map_size):
        self.map_size = map_size
        self.width = 10
        self.max_x = map_size
        self.max_y = map_size
        self.min_x = 0
        self.min_y = 0
        self.radius = 15

        self.init_pos, self.init_end = self.init_position()

        self.road_point = []

        self.current_pos = [self.init_pos, self.init_end]
        self.all_position_list = [[self.init_pos, self.init_end]]

    def init_position(self):
        """select a random initial position from the middle of
        one of the boundaries
        """
        option = rm.randint(0, 3)
        option = 1
        if option == 0:
            pos = np.array((self.max_x / 2, 5))
            end = np.array((pos[0] + self.width, pos[1]))
        elif option == 1:
            # pos = np.array((5, self.max_y / 2))
            pos = np.array((self.max_y / 2 - self.width / 2, self.max_y / 2))
            # end = np.array((pos[0], pos[1] - self.width))
            end = np.array((self.max_y / 2 + self.width / 2, self.max_y / 2))
        elif option == 2:
            pos = np.array((self.max_x / 2, self.max_y - 5))
            end = np.array((pos[0] + self.width, pos[1]))
        elif option == 3:
            pos = np.array((self.max_x - 5, self.max_y / 2))
            end = np.array((pos[0], pos[1] + self.width))

        return pos, end

    def position_to_line(self, position):
        x = [position[0][0], position[1][0], (position[0][0] + position[1][0]) / 2]
        y = [position[0][1], position[1][1], (position[0][1] + position[1][1]) / 2]
        return x, y

    def position_to_center(self):
        x = (self.current_pos[0][0] + self.current_pos[1][0]) / 2
        y = (self.current_pos[0][1] + self.current_pos[1][1]) / 2
        self.road_point = [x, y]
        # return [x, y]

    def point_in_range(self, a):
        """check if point is in the acceptable range"""
        # print("Point check", a)
        if 4 <= a[0] < (self.max_x - 4) and 4 <= a[1] < (self.max_y - 4):
            return 1
        else:
            return 0

    def point_in_range_2(self, a):
        """check if point is in the acceptable range"""
        if ((4) < a[0] and a[0] < (self.max_x - 4)) and ((4) <= a[1] and a[1] < (self.max_y - 4)):
            return 1
        else:
            return 0

    def go_straight(self, distance):
        a = self.current_pos[0]
        b = self.current_pos[1]

        test_distance = 1

        if self.point_in_range(a) == 0 or self.point_in_range(b) == 0:
            return False

        if (b - a)[1] > 0:
            p_a = b
            p_b = a
        elif (b - a)[1] < 0:
            p_a = a
            p_b = b
        elif (b - a)[1] == 0:
            if (b - a)[0] > 0:
                p_a = b
                p_b = a
            else:
                p_a = a
                p_b = b

        u_v = (p_a - p_b) / np.linalg.norm(p_b - p_a)
        sector = self.get_sector()

        if len(self.all_position_list) < 2:
            if sector == 0:
                R = np.array([[0, -1], [1, 0]])  # antilockwise
            elif sector == 1:
                R = np.array([[0, 1], [-1, 0]])  # clockwise
            elif sector == 2:
                R = np.array([[0, 1], [-1, 0]])
            elif sector == 3:
                R = np.array([[0, -1], [1, 0]])

            u_v_ = R.dot(u_v)

            p_a_ = p_a + u_v_ * distance
            p_b_ = p_b + u_v_ * distance

            self.current_pos = [p_a_, p_b_]
            self.all_position_list.append(self.current_pos)
            return True
        else:
            R = np.array([[0, -1], [1, 0]])
            u_v_ = R.dot(u_v)
            p_a_ = p_a + u_v_ * test_distance  # make a small perturbation
            p_b_ = p_b + u_v_ * test_distance

            new_pos = [p_a_, p_b_]
            if self.in_polygon(new_pos) == True:  # check if it's in correct direction
                R = np.array([[0, 1], [-1, 0]])
                u_v = R.dot(u_v)
                p_a_ = p_a + u_v * distance
                p_b_ = p_b + u_v * distance
                self.current_pos = [p_a_, p_b_]
                self.all_position_list.append(self.current_pos)
                return True
            else:
                p_a_ = p_a + u_v_ * distance
                p_b_ = p_b + u_v_ * distance
                self.current_pos = [p_a_, p_b_]
                self.all_position_list.append(self.current_pos)
                return True

    def turn_right(self, angle):
        a = self.current_pos[0]
        b = self.current_pos[1]
        test_angle = 3
        if self.point_in_range(a) == 0 or self.point_in_range(b) == 0:
            return False

        if (b - a)[1] > 0:
            p_a = b
            p_b = a
        elif (b - a)[1] < 0:
            p_a = a
            p_b = b
        elif (b - a)[1] == 0:
            if (b - a)[0] > 0:
                p_a = b
                p_b = a
            else:
                p_a = a
                p_b = b

        new_pos = self.clockwise_turn_top(test_angle, p_a, p_b)

        if self.in_polygon(new_pos) == True:
            self.current_pos = self.clockwise_turn_bot(angle, p_a, p_b)
        else:
            self.current_pos = self.clockwise_turn_top(angle, p_a, p_b)

        self.all_position_list.append(self.current_pos)
        return True

    def turn_left(self, angle):
        a = self.current_pos[0]
        b = self.current_pos[1]
        test_angle = 3
        if self.point_in_range(a) == 0 or self.point_in_range(b) == 0:
            return False

        if (b - a)[1] > 0:
            p_a = b
            p_b = a
        elif (b - a)[1] < 0:
            p_a = a
            p_b = b
        elif (b - a)[1] == 0:
            if (b - a)[0] > 0:
                p_a = b
                p_b = a
            else:
                p_a = a
                p_b = b

        new_pos = self.anticlockwise_turn_top(test_angle, p_a, p_b)

        if self.in_polygon(new_pos) == True:
            self.current_pos = self.anticlockwise_turn_bot(angle, p_a, p_b)
        else:
            self.current_pos = self.anticlockwise_turn_top(angle, p_a, p_b)
        self.all_position_list.append(self.current_pos)
        return True

    def clockwise_turn_top(self, angle, p_a, p_b):
        angle += 180
        radius = self.radius

        u_v = (p_a - p_b) / np.linalg.norm(p_a - p_b)
        o_o = p_a + u_v * radius

        o_b_norm = np.linalg.norm(o_o - p_b)

        o_a_norm = np.linalg.norm(o_o - p_a)

        o_b = (o_o - p_b) / o_b_norm
        o_a = (o_o - p_a) / o_a_norm

        R = np.array(
            [
                [np.cos(m.radians(angle)), np.sin(m.radians(angle))],
                [-np.sin(m.radians(angle)), np.cos(m.radians(angle))],
            ]
        )
        o_b_ = R.dot(o_b) * o_b_norm
        o_a_ = R.dot(o_a) * o_a_norm

        p_a_ = o_o + o_a_
        p_b_ = o_o + o_b_

        return [p_a_, p_b_]

    def clockwise_turn_bot(self, angle, p_a, p_b):
        radius = self.radius
        u_v = (p_a - p_b) / np.linalg.norm(p_a - p_b)
        o_o = p_b - u_v * radius
        o_b_norm = np.linalg.norm(o_o - p_b)
        o_a_norm = np.linalg.norm(o_o - p_a)
        o_b = (p_b - o_o) / o_b_norm
        o_a = (p_a - o_o) / o_a_norm

        R = np.array(
            [
                [np.cos(m.radians(angle)), np.sin(m.radians(angle))],
                [-np.sin(m.radians(angle)), np.cos(m.radians(angle))],
            ]
        )

        o_b_ = R.dot(o_b) * o_b_norm
        o_a_ = R.dot(o_a) * o_a_norm
        p_a_ = o_o + o_a_
        p_b_ = o_o + o_b_

        return [p_a_, p_b_]

    def anticlockwise_turn_top(self, angle, p_a, p_b):
        angle += 180
        radius = self.radius
        u_v = (p_a - p_b) / np.linalg.norm(p_a - p_b)
        o_o = p_a + u_v * radius

        o_b_norm = np.linalg.norm(o_o - p_b)

        o_a_norm = np.linalg.norm(o_o - p_a)

        o_b = (o_o - p_b) / o_b_norm
        o_a = (o_o - p_a) / o_a_norm

        R = np.array(
            [
                [np.cos(m.radians(angle)), -np.sin(m.radians(angle))],
                [np.sin(m.radians(angle)), np.cos(m.radians(angle))],
            ]
        )
        o_b_ = R.dot(o_b) * o_b_norm
        o_a_ = R.dot(o_a) * o_a_norm

        p_a_ = o_o + o_a_
        p_b_ = o_o + o_b_

        return [p_a_, p_b_]

    def anticlockwise_turn_bot(self, angle, p_a, p_b):
        radius = self.radius
        u_v = (p_a - p_b) / np.linalg.norm(p_a - p_b)
        o_o = p_b - u_v * radius

        o_b_norm = np.linalg.norm(o_o - p_b)
        o_a_norm = np.linalg.norm(o_o - p_a)
        o_b = (p_b - o_o) / o_b_norm
        o_a = (p_a - o_o) / o_a_norm

        R = np.array(
            [
                [np.cos(m.radians(angle)), -np.sin(m.radians(angle))],
                [np.sin(m.radians(angle)), np.cos(m.radians(angle))],
            ]
        )
        o_b_ = R.dot(o_b) * o_b_norm
        o_a_ = R.dot(o_a) * o_a_norm

        p_a_ = o_o + o_a_
        p_b_ = o_o + o_b_

        return [p_a_, p_b_]

    def in_polygon(self, new_position):
        """checks whether a point lies within a polygon
        between current and previous vector"""
        if len(self.all_position_list) <= 1:
            return True
        current = self.all_position_list[-1]
        prev = self.all_position_list[-2]
        new = new_position
        new_mid = (new[0] + new[1]) / 2

        point = Point(new_mid[0], new_mid[1])
        polygon = Polygon([tuple(current[0]), tuple(current[1]), tuple(prev[0]), tuple(prev[1])])
        return polygon.contains(point)

    def get_sector(self):
        """returns the sector of initial position"""

        return 1

    def remove_invalid_cases(self, points, states):
        points = list(points)
        new_list = [points[0]]
        new_states = {}
        # print("ALL points", points)
        i = 1
        while i < len(points):
            if self.point_in_range_2(points[i]) == 1:
                new_list.append(points[i])
                new_states["st" + str(i - 1)] = states["st" + str(i - 1)]
            else:
                return new_states, new_list
            i += 1

        return new_states, new_list

    def get_points_from_states(self, states):

        self.init_pos, self.init_end = self.init_position()
        self.current_pos = [self.init_pos, self.init_end]
        self.all_position_list = [[self.init_pos, self.init_end]]

        self.position_to_center()
        points = [self.road_point]
        tc = states
        for state in tc:
            action = tc[state]["state"]
            if action == "straight":
                if self.go_straight(tc[state]["value"]) == True:
                    self.position_to_center()
                    point = self.road_point
                    points.append(point)
            elif action == "left":
                if self.turn_left(tc[state]["value"]) == True:
                    self.position_to_center()
                    point = self.road_point
                    points.append(point)
            elif action == "right":
                if self.turn_right(tc[state]["value"]) == True:
                    self.position_to_center()
                    point = self.road_point
                    points.append(point)
            else:
                print("ERROR")
        return points

    def build_tc(self, points):

        time_ = str(int(time.time()))

        fig, ax = plt.subplots(figsize=(12, 12))
        # , nodes[closest_index][0], nodes[closest_index][1], 'go'
        road_x = []
        road_y = []
        for p in points:
            road_x.append(p[0])
            road_y.append(p[1])

        ax.plot(road_x, road_y, "yo--", label="Road")

        top = self.map_size
        bottom = 0

        ax.set_title("Test case fitenss ", fontsize=17)

        ax.set_ylim(bottom, top)

        ax.set_xlim(bottom, top)

        fig.savefig(".\\Test\\" + time_ + "+test.jpg")
        ax.legend()
        plt.close(fig)


def read_schedule(tc, size):
    time_ = str(int(time.time()))
    fig, ax1 = plt.subplots(figsize=(12, 12))
    car_map = Map(size)
    for state in tc:
        action = tc[state]["state"]
        print("location:", car_map.current_pos)
        if action == "straight":
            car_map.go_straight(tc[state]["value"])
            print(tc[state]["value"])
            x, y = car_map.position_to_line(car_map.current_pos)
            ax1.plot(x, y, "o--y")
        elif action == "left":
            car_map.turn_left(tc[state]["value"])
            print(tc[state]["value"])
            x, y = car_map.position_to_line(car_map.current_pos)
            ax1.plot(x, y, "o--y")
        elif action == "right":
            car_map.turn_right(tc[state]["value"])
            print(tc[state]["value"])
            x, y = car_map.position_to_line(car_map.current_pos)
            ax1.plot(x, y, "o--y")
        else:
            print("Wrong value")

    top = size
    bottom = 0
    ax1.set_ylim(bottom, top)
    ax1.set_xlim(bottom, top)
    fig.savefig(".\\Test\\" + time_ + "+test2.jpg")
    plt.close(fig)
